import logging
import httpx
import json
import time
import stamina

client = httpx.Client(timeout=10)
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.NullHandler())


def buildCommand(CompId, cmd, params={}, equiv=None):
    res = {"cid": CompId, "cmd": cmd, "params": params, "equiv": equiv}

    return res


def buildRequestBody(sessionId, windowId, pageId, cmdPipe, mode="ajax"):
    cmdPipeFill = {"queue": [{"wid": windowId, "pid": pageId, **d} for d in cmdPipe]}
    cmdJson = json.dumps(cmdPipeFill)
    now = time.time()
    timestamp = str(int(now * 1000))

    res = {
        "sid": sessionId,
        "wid": windowId,
        "pid": pageId,
        "data": cmdJson,
        "ts": timestamp,
        "mode": mode,
    }

    return res


def buildBrowseCommand(credentials, ident, value):
    # type check
    if ident in ["type", "HA", "NA", "host"]:
        assert isinstance(value, list), f"{ident} must be a list"
        value = [str(v) for v in value]
    elif ident in ["collectDateFrom", "collectDateTo"]:
        assert isinstance(value, str), f"{ident} must be a string"

    # host code
    if ident == "host":
        value = [credentials.hostCode[v.title()] for v in value]

    identCeid = credentials.browseParamsCeid[ident]

    res = [
        buildCommand(
            CompId=credentials.browsePage["browseFormCompId"],
            cmd="setTarget",
            params={"cvalue": value, "ceid": identCeid},
            equiv=f"ST{identCeid}",
        ),
        buildCommand(
            CompId=credentials.browsePage["browseFormCompId"],
            cmd="ChangeValue",
            params={"cvalue": value, "ceid": identCeid},
            equiv=f"CV{identCeid}",
        ),
        buildCommand(
            CompId=credentials.browsePage["browseFormCompId"],
            cmd=credentials.browseParamsCmd[ident],
            params={"ceid": identCeid},
        ),
    ]

    return res


def buildBatch(startIdx, endIdx, batchSize):
    res = []
    for i in range((endIdx - startIdx) // batchSize + 1):
        start = i * batchSize
        if i != (endIdx - startIdx) // batchSize:
            end = i * batchSize + batchSize - 1
        else:
            end = endIdx
        count = end - start + 1
        res.append({"start": start, "end": end, "count": count})

    return res


################## requests ####################


@stamina.retry(on=httpx.HTTPError, attempts=3)
def httpGet(url, headers):
    res = client.get(url, headers=headers, follow_redirects=True)
    return res


@stamina.retry(on=httpx.HTTPError, attempts=3)
def httpPost(url, data, headers):
    res = client.post(url, data=data, headers=headers, follow_redirects=True)
    return res


################## page ####################


def resultToBrowsePage(credentials):
    cmdPipe = [
        buildCommand(CompId=credentials.resultPage["downloadCompId"], cmd="GoBack")
    ]
    body = buildRequestBody(
        credentials.sessionId,
        credentials.windowId,
        credentials.resultPage["pid"],
        cmdPipe,
    )

    httpPost(credentials.url, data=body, headers=credentials.headers)

    browsePagePid = credentials.browsePage["pid"]
    httpGet(
        f"{credentials.url}?sid={credentials.sessionId}&pid={browsePagePid}",
        headers=credentials.headers,
    )

    return None


def downloadToResultPage(credentials):
    cmdPipe = [
        buildCommand(
            CompId=credentials.downloadPage["resultDownloadCompId"], cmd="Cancel"
        ),
    ]

    body = buildRequestBody(
        credentials.sessionId,
        credentials.downloadWindowId,
        credentials.downloadPage["pid"],
        cmdPipe,
    )

    httpPost(credentials.url, data=body, headers=credentials.headers)
    resultPagePid = credentials.resultPage["pid"]
    httpGet(
        f"{credentials.url}?sid={credentials.sessionId}&pid={resultPagePid}",
        headers=credentials.headers,
    )

    return None


################## logger ####################
def log(level=logging.DEBUG):
    logger = logging.getLogger(__package__)
    handler = logging.StreamHandler()
    handler.setLevel(level)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    return None



import pytest
import gisflu


@pytest.mark.parametrize("passwd", ["12345678", "abcdefdfs", "as52345fasdf4"])
def test_passwd_length(passwd):
    assert len(passwd) >= 8


def test_login():
    cred = gisflu.login()
    assert cred.sessionId is not None, "Failed to fetch GISAID"
    assert cred.browseParamsCeid["type"] is not None, "Failed to parse browse page"
    assert (
        cred.downloadParamsCeid["downloadConfirm"] is not None
    ), "Failed to parse download page"



import re
from .utils import (
    buildCommand,
    buildRequestBody,
    buildBrowseCommand,
    buildBatch,
    resultToBrowsePage,
    httpGet,
    httpPost,
)
from .credentials import credentials
from tqdm import tqdm
import pandas as pd
import logging
from typing import List

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.NullHandler())


def search(
    cred: credentials,
    type: List[str] | None = None,
    HA: List[str] | None = None,
    NA: List[str] | None = None,
    host: List[str] | None = None,
    collectDateFrom: str | None = None,
    collectDateTo: str | None = None,
    recordLimit: int = 50,
) -> pd.DataFrame:
    """
    Search for records in the GISAID Flu database based on specified criteria.

    Args:
        cred (credentials): The credentials object containing session information.
        type (List[str], optional): A list of virus types to filter the search results. Defaults to None.
        HA (List[str], optional): A list of hemagglutinin (HA) subtypes to filter the search results. Defaults to None.
        NA (List[str], optional): A list of neuraminidase (NA) subtypes to filter the search results. Defaults to None.
        host (List[str], optional): A list of host species to filter the search results. Defaults to None.
        collectDateFrom (str, optional): The starting date for the collection date filter. Defaults to None.
        collectDateTo (str, optional): The ending date for the collection date filter. Defaults to None.
        recordLimit (int, optional): The maximum number of records to return. Defaults to 50.

    Return:
        pd.DataFrame: A DataFrame containing the search results.

    Example:
        ```
        cred = gisflu.login()
        gisflu.search(cred, type=["A"], HA=["3"], NA=["2"],
            collectDateFrom="2020-01-01", recordLimit=10)
        ```
    """

    # search by command pipeline
    cmdPipe = []
    if type:
        cmdPipe += buildBrowseCommand(cred, "type", type)
    if HA:
        cmdPipe += buildBrowseCommand(cred, "HA", HA)
    if NA:
        cmdPipe += buildBrowseCommand(cred, "NA", NA)
    if host:
        cmdPipe += buildBrowseCommand(cred, "host", host)
    if collectDateFrom:
        cmdPipe += buildBrowseCommand(cred, "collectDateFrom", collectDateFrom)
    if collectDateTo:
        cmdPipe += buildBrowseCommand(cred, "collectDateTo", collectDateTo)

    body = buildRequestBody(
        cred.sessionId, cred.windowId, cred.browsePage["pid"], cmdPipe
    )

    res = httpPost(cred.url, data=body, headers=cred.headers)

    # records count in the browse page
    preResultText = res.text

    recordCount, recordSeqCount = [
        int(i.replace(",", ""))
        for i in re.search(
            r"Total: ([\d,]+) viruses \(([\d,]+) sequences\)", preResultText
        ).group(1, 2)
    ]

    logger.info(f"{recordCount} records, {recordSeqCount} seqs found")

    # refresh result page id
    cmdPipe = [buildCommand(CompId=cred.browsePage["searchButtonCompId"], cmd="search")]
    body = buildRequestBody(
        cred.sessionId, cred.windowId, cred.browsePage["pid"], cmdPipe
    )
    res = httpPost(cred.url, data=body, headers=cred.headers)
    resultPagePid = re.search(r"sys.goPage\(\'(.+?)\'\)", res.text).group(1)
    cred.resultPage["pid"] = resultPagePid

    logger.debug("Parse result page...")
    # go to result page
    res = httpGet(
        f"{cred.url}?sid={cred.sessionId}&pid={resultPagePid}", headers=cred.headers
    )
    resultPageText = res.text
    cred.resultPage["resultCompId"] = re.search(
        r"sys\.createComponent\(\'(c_\w+?)\',\'IsolateResultListComponent\'",
        resultPageText,
    ).group(1)

    logger.debug("Fetch result records...")
    # fetch records
    if recordCount > 0:
        resultJson = []

        batches = buildBatch(0, min(recordCount, recordLimit) - 1, batchSize=27)
        for batch in tqdm(batches):
            cmdPipe = [
                buildCommand(
                    CompId=cred.resultPage["resultCompId"],
                    cmd="SetPaginating",
                    params={
                        "start_index": batch["start"],
                        "rows_per_page": batch["count"],
                    },
                ),
                buildCommand(CompId=cred.resultPage["resultCompId"], cmd="GetData"),
            ]

            body = buildRequestBody(
                cred.sessionId, cred.windowId, cred.resultPage["pid"], cmdPipe
            )
            res = httpPost(cred.url, data=body, headers=cred.headers)

            resultJson += res.json()["records"]

        # records dataframe
        reslutDF = pd.DataFrame(resultJson)

        reslutDF = reslutDF.drop(
            [s for s in reslutDF.columns if s not in cred.resultHeaderDict.keys()],
            axis=1,
        )

        reslutDF = reslutDF.rename(columns=cred.resultHeaderDict)

        reslutDF = reslutDF.drop(["__toggle__", "edit", "HE", "P3"], axis=1)

        for col in ["Name", "PB2", "PB1", "PA", "HA", "NP", "NA", "MP", "NS"]:
            reslutDF[col] = reslutDF[col].str.replace(
                r"^.+?>(.+?)</.+$", r"\1", regex=True
            )
    else:
        reslutDF = pd.DataFrame()

    resultToBrowsePage(cred)

    nrow = reslutDF.shape[0]
    logger.debug(f"Search completed: return {nrow} rows")

    return reslutDF



from .login import login
from .utils import log
from .browse import search
from .download import download
from dotenv import load_dotenv

load_dotenv()


__all__ = ["log", "login", "search", "download"]



import re
from .utils import (
    buildCommand,
    buildRequestBody,
    httpGet,
    httpPost,
    downloadToResultPage,
    resultToBrowsePage,
)
from .credentials import credentials
import logging
from datetime import datetime
from typing import List
import urllib

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.NullHandler())


def download(
    cred: credentials,
    isolateIds: List[str],
    downloadType: str = "protein",
    segments: List[str] = ["HA", "NA"],
    filename: str | None = None,
) -> None:
    """
    Downloads records for the given isolate IDs.

    Args:
        cred (object): The credentials object.
        isolateIds (list): List of isolate IDs to download data for.
        downloadType (str, optional): The type of data to download. Defaults to "protein".
        segments (list, optional): List of segments to download. Defaults to ["HA", "NA"].
        filename (str, optional): The name of the file to save the downloaded data. If not provided, a default filename will be generated.

    Return:
        None

    Example:
        ```
        cred = gisflu.login()
        isolateIds = ["EPI_ISL_19185107", "EPI_ISL_19151100"]
        gisflu.download(cred, isolateIds, downloadType="protein", segments=["HA", "NA"],
            filename="records.fasta")
        ```
    """

    assert all(
        id.startswith("EPI_ISL_") for id in isolateIds
    ), 'isolateId must start with "EPI_ISL_"'

    assert downloadType in [
        "metadata",
        "protein",
        "dna",
    ], "downloadType must be metadata|protein|dna"

    segmentCheck = [
        "NP",
        "P3",
        "HA",
        "M1",
        "M2",
        "BM2",
        "CM2",
        "M",
        "NA",
        "NB",
        "NS1",
        "NEP",
        "NS2",
        "PA",
        "PA-X",
        "PB1-F2",
        "PB1",
        "HE",
        "PB2",
    ]

    unknownSegments = [segment for segment in segments if segment not in segmentCheck]
    unknownSegmentStr = ", ".join(unknownSegments)
    assert len(unknownSegments) == 0, f"Unknown segment(s): {unknownSegmentStr}"

    logger.debug("Go to result page...")
    # fetch result page id
    cmdPipe = [buildCommand(CompId=cred.browsePage["searchButtonCompId"], cmd="search")]
    body = buildRequestBody(
        cred.sessionId, cred.windowId, cred.browsePage["pid"], cmdPipe
    )
    res = httpPost(cred.url, data=body, headers=cred.headers)
    resultPagePid = re.search(r"sys.goPage\(\'(.+?)\'\)", res.text).group(1)
    cred.resultPage["pid"] = resultPagePid

    # go to result page
    res = httpGet(
        f"{cred.url}?sid={cred.sessionId}&pid={resultPagePid}", headers=cred.headers
    )

    # select records, get download page id
    cmdPipe = [
        buildCommand(
            CompId=cred.resultPage["resultCompId"],
            cmd="ChangeValue",
            params={
                "row_id": acc.replace("EPI_ISL_", ""),
                "col_name": "c",
                "value": True,
            },
        )
        for acc in isolateIds
    ]
    cmdPipe += [buildCommand(CompId=cred.resultPage["downloadCompId"], cmd="Download")]

    body = buildRequestBody(
        cred.sessionId, cred.windowId, cred.resultPage["pid"], cmdPipe
    )

    res = httpPost(cred.url, data=body, headers=cred.headers)

    cred.downloadWindowId, cred.downloadPage["pid"] = re.search(
        r"sys.openOverlay\(\'(\w+?)\',\'(\w+?)\'", res.text
    ).group(1, 2)

    logger.debug("Go to download page...")
    # go to download overlay page
    res = httpGet(
        f'{cred.url}?sid={cred.sessionId}&pid={cred.downloadPage["pid"]}',
        headers=cred.headers,
    )
    resultDownloadCompId = cred.downloadPage["resultDownloadCompId"]

    logger.debug("Set download params...")
    if downloadType == "metadata":
        cmdPipe = [
            buildCommand(CompId=resultDownloadCompId, cmd="download"),
        ]

        body = buildRequestBody(
            cred.sessionId, cred.downloadWindowId, cred.downloadPage["pid"], cmdPipe
        )

        res = httpPost(cred.url, data=body, headers=cred.headers)

        api = re.search(r"sys\.downloadFile\(\\\"(.+?)\\\"", res.text).group(1)
    elif downloadType in ["protein", "dna"]:
        if downloadType == "protein":
            typeCvalue = "proteins"
            downloadSegmentCeid = cred.downloadParamsCeid["proteinSegment"]
            faHeader = "Protein Accession no.|Gene name|Isolate name|Isolate ID|Type@Collection date"
        else:
            typeCvalue = "dna"
            downloadSegmentCeid = cred.downloadParamsCeid["dnaSegment"]
            faHeader = (
                "DNA Accession no.|Segment|Isolate name|Isolate ID|Type@Collection date"
            )

        resultDownloadCompId = cred.downloadPage["resultDownloadCompId"]
        downloadFormatCeid = cred.downloadParamsCeid["downloadFormat"]
        fastaHeaderCeid = cred.downloadParamsCeid["fastaHeader"]

        cmdPipe = [
            # select protein|dna
            buildCommand(
                CompId=resultDownloadCompId,
                cmd="setTarget",
                params={
                    "cvalue": typeCvalue,
                    "ceid": downloadFormatCeid,
                },
                equiv=f"ST{downloadFormatCeid}",
            ),
            buildCommand(
                CompId=resultDownloadCompId,
                cmd="ChangeValue",
                params={
                    "cvalue": typeCvalue,
                    "ceid": downloadFormatCeid,
                },
                equiv=f"CV{downloadFormatCeid}",
            ),
            buildCommand(
                CompId=resultDownloadCompId,
                cmd="ShowProteins",
                params={"ceid": downloadFormatCeid},
            ),
            # check segment
            buildCommand(
                CompId=resultDownloadCompId,
                cmd="setTarget",
                params={"cvalue": segments, "ceid": downloadSegmentCeid},
                equiv=f"ST{downloadSegmentCeid}",
            ),
            buildCommand(
                CompId=resultDownloadCompId,
                cmd="ChangeValue",
                params={"cvalue": segments, "ceid": downloadSegmentCeid},
                equiv=f"CV{downloadSegmentCeid}",
            ),
            buildCommand(
                CompId=resultDownloadCompId,
                cmd="SelChange",
                params={"ceid": downloadSegmentCeid},
            ),
            # set fasta header
            buildCommand(
                CompId=resultDownloadCompId,
                cmd="setTarget",
                params={"cvalue": faHeader, "ceid": fastaHeaderCeid},
                equiv=f"ST{fastaHeaderCeid}",
            ),
            buildCommand(
                CompId=resultDownloadCompId,
                cmd="ChangeValue",
                params={"cvalue": faHeader, "ceid": fastaHeaderCeid},
                equiv=f"CV{fastaHeaderCeid}",
            ),
            buildCommand(
                CompId=resultDownloadCompId,
                cmd="fillExampleCopied",
                params={"ceid": fastaHeaderCeid},
            ),
            # download
            buildCommand(CompId=resultDownloadCompId, cmd="download"),
        ]

        body = buildRequestBody(
            cred.sessionId, cred.downloadWindowId, cred.downloadPage["pid"], cmdPipe
        )

        res = httpPost(cred.url, data=body, headers=cred.headers)

        api = re.search(r"sys\.downloadFile\(\\\"(.+?)\\\"", res.text).group(1)

    # download
    logger.debug("Downloading...")
    now = datetime.now().strftime("%Y%m%d-%H%M%S")
    count = len(isolateIds)
    if filename is None:
        if downloadType == "metadata":
            extension = "xls"
        elif downloadType in ["protein", "dna"]:
            extension = "fasta"
        filename = f"gisflu-{downloadType}-{count}records-{now}.{extension}"

    downloadLink = "https://" + urllib.parse.urlparse(cred.url).hostname + api
    res = httpGet(downloadLink, headers=cred.headers)

    with open(filename, "wb") as f:
        f.write(res.content)

    downloadToResultPage(cred)
    resultToBrowsePage(cred)

    return None



class credentials:
    def __init__(self):
        self.url = "https://platform.epicov.org/epi3/frontend"
        self.headers = {
            "accept": "application/json, text/javascript, */*; q=0.01",
            "content-type": "application/x-www-form-urlencoded; charset=UTF-8",
        }
        self.sessionId = None
        self.windowId = None
        self.downloadWindowId = None
        self.loginPage = {"pid": None, "loginCompId": None}
        self.firstPage = {"pid": None, "dbSwitchCompId": None}
        self.homePage = {"pid": None, "browseCompId": None}
        self.browsePage = {
            "pid": None,
            "browseFormCompId": None,
            "searchButtonCompId": None,
        }
        self.resultPage = {
            "pid": None,
            "resultCompId": None,
            "downloadCompId": None,
        }
        self.downloadPage = {"pid": None, "resultDownloadCompId": None}
        self.browseParamsCeid = {}
        self.browseParamsCmd = {
            **{i: "TypeChanged" for i in ["type", "HA", "NA"]},
            **{i: "OnlyCount" for i in ["host", "collectDateFrom", "collectDateTo"]},
        }
        self.hostCode = {"Human": "101", "Animal": "102", "Avian": "103"}
        self.resultHeaderDict = {}
        self.downloadParamsCeid = {}

    def __repr__(self):
        return f"credentials(sid={self.sessionId})"



import os
import hashlib
import re
from .credentials import credentials
from .utils import (
    buildCommand,
    buildRequestBody,
    resultToBrowsePage,
    downloadToResultPage,
    httpGet,
    httpPost,
)
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.NullHandler())


def login(username: str | None = None, password: str | None = None) -> credentials:
    """
    Login the GISAID Flu database, parse elements ids and store them in a credentials object.

    Args:
        username (str, optional): The username to log in with. If not provided, it will be fetched from the environment variable "GISAID_USERNAME".
        password (str, optional): The password to log in with. If not provided, it will be fetched from the environment variable "GISAID_PASSWORD".

    Return:
        credentials

    Example:
        ```
        # Log in with provided username and password
        gisflu.login("myusername", "mypassword")

        # Log in using environment variables
        gisflu.login()
        ```
    """

    cred = credentials()

    # get username and password
    if username is None or password is None:
        logger.debug(
            "Username and password not provided, fetching from environment variables"
        )

        username = os.getenv("GISAID_USERNAME")
        password = os.getenv("GISAID_PASSWORD")

        assert (
            username is not None
        ), 'Please set the environment variable "GISAID_USERNAME"'
        assert (
            password is not None
        ), 'Please set the environment variable "GISAID_PASSWORD"'

    password_md5 = hashlib.md5(password.encode()).hexdigest()

    # fetch sessionId first
    res = httpGet(cred.url, headers=cred.headers)
    cred.sessionId = re.search(r'name="sid" value=\'(.+?)\'', res.text).group(1)
    logger.debug(f"Get sessionId: {cred.sessionId}")

    # then get login page, to get more ids
    res = httpGet(f"{cred.url}?sid={cred.sessionId}", headers=cred.headers)
    loginPageText = res.text
    cred.windowId = re.search(r'sys\["WID"\] = "(.+?)";', loginPageText).group(1)
    cred.loginPage["pid"] = re.search(r'sys\["PID"\] = "(.+?)";', loginPageText).group(
        1
    )
    cred.loginPage["loginCompId"] = re.search(
        r"sys.getC\(\'(.+?)\'\).call\(\'doLogin\'", loginPageText
    ).group(1)

    # login by command pipeline
    cmdPipe = [
        buildCommand(
            CompId=cred.loginPage["loginCompId"],
            cmd="doLogin",
            params={"login": username, "hash": password_md5},
        )
    ]

    body = buildRequestBody(
        cred.sessionId, cred.windowId, cred.loginPage["pid"], cmdPipe, mode="ajax"
    )

    res = httpPost(cred.url, data=body, headers=cred.headers)
    assert re.search("cms_page", res.text), "Username or password wrong!"
    logger.debug("username and password validated!")

    # first page after login
    logger.debug("Go to first page...")
    res = httpGet(f"{cred.url}?sid={cred.sessionId}", headers=cred.headers)
    firstPageText = res.text
    cred.firstPage["pid"] = re.search(r'sys\["PID"\] = "(.+?)";', firstPageText).group(
        1
    )
    cred.firstPage["dbSwitchCompId"] = re.search(
        r"sys.call\(\'(.+?)\',\'Go\'", firstPageText
    ).group(1)

    # fetch flu home page id by command pipeline
    logger.debug("Go to flu homepage...")
    cmdPipe = [
        buildCommand(
            CompId=cred.firstPage["dbSwitchCompId"], cmd="Go", params={"page": "epi3"}
        )
    ]

    body = buildRequestBody(
        cred.sessionId, cred.windowId, cred.firstPage["pid"], cmdPipe
    )

    res = httpPost(cred.url, data=body, headers=cred.headers)
    homePagePid = re.search(r"sys.goPage\(\'(.+?)\'\)", res.text).group(1)
    cred.homePage["pid"] = homePagePid

    # go to flu home page
    res = httpGet(
        f"{cred.url}?sid={cred.sessionId}&pid={homePagePid}", headers=cred.headers
    )
    homePageText = res.text

    ################## browse page ####################
    logger.debug("Parse browse page...")

    # fetch browse(search) page id
    cred.homePage["browseCompId"] = re.search(
        r"class=\"sys-actionbar-action-ni\" onclick=\"sys.getC\(\'(.+?)\'\)",
        homePageText,
    ).group(1)

    cmdPipe = [buildCommand(CompId=cred.homePage["browseCompId"], cmd="Browse")]

    body = buildRequestBody(
        cred.sessionId, cred.windowId, cred.homePage["pid"], cmdPipe
    )

    res = httpPost(cred.url, data=body, headers=cred.headers)

    browsePagePid = re.search(r"sys.goPage\(\'(.+?)\'\)", res.text).group(1)
    cred.browsePage["pid"] = browsePagePid

    # go to browse page
    res = httpGet(
        f"{cred.url}?sid={cred.sessionId}&pid={browsePagePid}", headers=cred.headers
    )
    browsePageText = res.text

    cred.browsePage["browseFormCompId"] = re.search(
        r"sys\.createComponent\(\'(c_\w+?)\',\'IsolateBrowseFormComponent\'",
        browsePageText,
    ).group(1)

    cred.browsePage["searchButtonCompId"] = re.search(
        r"sys\.createComponent\(\'(c_\w+?)\',\'IsolateSearchButtonsComponent\'",
        browsePageText,
    ).group(1)

    # fetch browse component event id
    browseItemText = re.findall(r"createFI\(.+?function", browsePageText)

    browseItemDict = {}
    for s in browseItemText:
        ident = re.search(r"Widget\',\'(.+?)\',function", s).group(1)
        ceid = re.search(r"createFI\(\'(.+?)\',", s).group(1)
        browseItemDict[ident] = ceid

    cred.browseParamsCeid["type"] = browseItemDict["isl_type"]
    cred.browseParamsCeid["HA"] = browseItemDict["isl_subtype_h"]
    cred.browseParamsCeid["NA"] = browseItemDict["isl_subtype_n"]
    cred.browseParamsCeid["lineage"] = browseItemDict["isl_lineage"]
    cred.browseParamsCeid["host"] = browseItemDict["isl_host"]
    cred.browseParamsCeid["location"] = browseItemDict["isl_location"]
    cred.browseParamsCeid["collectDateFrom"] = browseItemDict["isl_collect_date_from"]
    cred.browseParamsCeid["collectDateTo"] = browseItemDict["isl_collect_date_to"]

    ################## result page ####################
    logger.debug("Parse result page...")

    # fetch result page id
    cmdPipe = [buildCommand(CompId=cred.browsePage["searchButtonCompId"], cmd="search")]
    body = buildRequestBody(
        cred.sessionId, cred.windowId, cred.browsePage["pid"], cmdPipe
    )
    res = httpPost(cred.url, data=body, headers=cred.headers)
    resultPagePid = re.search(r"sys.goPage\(\'(.+?)\'\)", res.text).group(1)
    cred.resultPage["pid"] = resultPagePid

    # go to result page
    res = httpGet(
        f"{cred.url}?sid={cred.sessionId}&pid={resultPagePid}", headers=cred.headers
    )
    resultPageText = res.text
    cred.resultPage["resultCompId"] = re.search(
        r"sys\.createComponent\(\'(c_\w+?)\',\'IsolateResultListComponent\'",
        resultPageText,
    ).group(1)
    cred.resultPage["downloadCompId"] = re.search(
        r"sys\.createComponent\(\'(c_\w+?)\',\'IsolateDownloadButtonComponent\'",
        resultPageText,
    ).group(1)

    # parse result table header
    tableHeaderText = re.findall(r"new Object\(\{\'label.+?cid", resultPageText)

    for s in tableHeaderText:
        label = re.search(r"label\':\'([\w ]+?)\'", s).group(1)
        key = re.search(r"key\':\'(\w+?)\'", s).group(1)
        cred.resultHeaderDict[key] = label

    ################## download page ####################
    logger.debug("Parse download page...")

    # get a temp record
    cmdPipe = [
        buildCommand(
            CompId=cred.resultPage["resultCompId"],
            cmd="SetPaginating",
            params={"start_index": 0, "rows_per_page": 27},
        ),
        buildCommand(CompId=cred.resultPage["resultCompId"], cmd="GetData"),
    ]

    body = buildRequestBody(
        cred.sessionId, cred.windowId, cred.resultPage["pid"], cmdPipe
    )
    res = httpPost(cred.url, data=body, headers=cred.headers)

    tempRecordId = res.json()["records"][0]["b"]

    # select this temp record
    cmdPipe = [
        buildCommand(
            CompId=cred.resultPage["resultCompId"],
            cmd="ChangeValue",
            params={"row_id": tempRecordId, "col_name": "c", "value": True},
        ),
        buildCommand(CompId=cred.resultPage["downloadCompId"], cmd="Download"),
    ]

    body = buildRequestBody(
        cred.sessionId, cred.windowId, cred.resultPage["pid"], cmdPipe
    )

    res = httpPost(cred.url, data=body, headers=cred.headers)

    cred.downloadWindowId, cred.downloadPage["pid"] = re.search(
        r"sys.openOverlay\(\'(\w+?)\',\'(\w+?)\'", res.text
    ).group(1, 2)

    # go to download overlay page
    res = httpGet(
        f'{cred.url}?sid={cred.sessionId}&pid={cred.downloadPage["pid"]}',
        headers=cred.headers,
    )
    downloadPageText = res.text

    cred.downloadPage["resultDownloadCompId"] = re.search(
        r"sys\.createComponent\(\'(c_\w+?)\',\'IsolateResultDownloadComponent\'",
        downloadPageText,
    ).group(1)

    # fetch download item ceid
    downloadItemText = re.findall(r"createFI\(.+?function", downloadPageText)

    downloadItemDict = {}
    for s in downloadItemText:
        ident = re.search(r"Widget\',\'(.+?)\',function", s).group(1)
        ceid = re.search(r"createFI\(\'(.+?)\',", s).group(1)
        downloadItemDict[ident] = ceid

    cred.downloadParamsCeid["downloadFormat"] = downloadItemDict["format"]
    cred.downloadParamsCeid["downloadConfirm"] = downloadItemDict["download"]

    # fetch protein segment ceid

    cmdPipe = [
        buildCommand(
            CompId=cred.downloadPage["resultDownloadCompId"],
            cmd="setTarget",
            params={
                "cvalue": "proteins",
                "ceid": cred.downloadParamsCeid["downloadFormat"],
            },
            equiv=f'ST{cred.downloadParamsCeid["downloadFormat"]}',
        ),
        buildCommand(
            CompId=cred.downloadPage["resultDownloadCompId"],
            cmd="ChangeValue",
            params={
                "cvalue": "proteins",
                "ceid": cred.downloadParamsCeid["downloadFormat"],
            },
            equiv=f'CV{cred.downloadParamsCeid["downloadFormat"]}',
        ),
        buildCommand(
            CompId=cred.downloadPage["resultDownloadCompId"],
            cmd="ShowProteins",
            params={"ceid": cred.downloadParamsCeid["downloadFormat"]},
        ),
    ]

    body = buildRequestBody(
        cred.sessionId, cred.downloadWindowId, cred.downloadPage["pid"], cmdPipe
    )

    res = httpPost(cred.url, data=body, headers=cred.headers)

    downloadProteinText = res.text

    cred.downloadParamsCeid["proteinSegment"] = re.search(
        r"createFI\(\'(\w+?)\',\'CheckboxWidget\',\'proteins\'", downloadProteinText
    ).group(1)

    # fetch dna segment ceid
    cmdPipe = [
        buildCommand(
            CompId=cred.downloadPage["resultDownloadCompId"],
            cmd="setTarget",
            params={
                "cvalue": "dna",
                "ceid": cred.downloadParamsCeid["downloadFormat"],
            },
            equiv=f'ST{cred.downloadParamsCeid["downloadFormat"]}',
        ),
        buildCommand(
            CompId=cred.downloadPage["resultDownloadCompId"],
            cmd="ChangeValue",
            params={
                "cvalue": "dna",
                "ceid": cred.downloadParamsCeid["downloadFormat"],
            },
            equiv=f'CV{cred.downloadParamsCeid["downloadFormat"]}',
        ),
        buildCommand(
            CompId=cred.downloadPage["resultDownloadCompId"],
            cmd="ShowProteins",
            params={"ceid": cred.downloadParamsCeid["downloadFormat"]},
        ),
    ]

    body = buildRequestBody(
        cred.sessionId, cred.downloadWindowId, cred.downloadPage["pid"], cmdPipe
    )

    res = httpPost(cred.url, data=body, headers=cred.headers)

    downloadDNAText = res.text

    cred.downloadParamsCeid["dnaSegment"] = re.search(
        r"createFI\(\'(\w+?)\',\'CheckboxWidget\',\'dna\'", downloadDNAText
    ).group(1)

    cred.downloadParamsCeid["fastaHeader"] = re.search(
        r"createFI\(\'(\w+?)\',\'EntryWidget\',\'header\'", downloadDNAText
    ).group(1)

    ################## return browse page ####################
    downloadToResultPage(cred)
    resultToBrowsePage(cred)
    logger.debug(f"{username} logged!")

    return cred
